#include<stddef.h>
#include<stdio.h>
#include<stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>           // mmap() is defined in this header
#include <fcntl.h>
#include <unistd.h>
#include <algorithm>

#define VERBOSE 0
#define APROXIMO 1

struct Pixel {
  int r;
  int g;
  int b;
  bool operator==(const Pixel &other) const {
    return (r == other.r) && (g == other.g) && (b == other.b);
  }
  bool operator!=(const Pixel &other) const {
    return (r != other.r) || (g != other.g) || (b != other.b);
  }
};

struct PatchNode {
  int   image;
  int   count;
  PatchNode *next;
};

struct PatchHolder {
  Pixel     p;
  PatchNode *nodes;
};

struct DistInd {
  int dist,ind;
  bool operator<(const DistInd &other) const {
    return dist<other.dist || (dist==other.dist && ind<other.ind);
  }
};

struct Contest {
  int la_k,W,bitsq;
  int newPatchNode;
  char *r; // para recorrer imagenes
  int width,height; // para cargar imagenes
  int numImages;
  static const int maxPixels     = 5000000;
  static const int perfectSize   =  438976; // 76**3
  static const int hashSize      = 1<<22; //1<<23;
  static const int hashMask      = hashSize-1;
  static const int maxpatchnodes = 5000000;
  union {
    PatchHolder hashVector[hashSize];
    PatchNode   *perfectTable[perfectSize];
  };
  PatchNode    patchNodeVector[maxpatchnodes];
  static const int maxImages = 1010;
  DistInd vec_selection[maxImages];
  int total_count[maxImages];
  // para calcular la matriz de costes
  static const int maxVFloyd = maxImages;
  static const int maxSFloyd = (maxVFloyd*(maxVFloyd-1))/2;
  int  *m[maxVFloyd];   // para indexar matriz triangular
  char *m_used[maxVFloyd];   // para indexar matriz triangular
  int   mat[maxSFloyd]; // datos matriz triangular
  char  mat_used[maxSFloyd]; // matriz booleanos para borrar aristas
  Pixel imagePixels[maxPixels];

  //----------------------------------------------------------------
  // para calcular las distancias

  inline int& eme(int u, int v) { // acceder a m de modo general pero
				  // ineficiente
    return (u>v) ? m[u][v] : m[v][u];
  }

  static const int INFINITE  = 100000000;
  static const int maxEdges  = maxImages*1000;
  static const int queueSize = maxImages*1000;
  int  vertex2edge[maxImages][2];
  bool stamp[maxImages];
  int queue[queueSize];
  int edges[maxEdges];

  //----------------------------------------------------------------
  inline int  readValue() {
    while (*r < '0') r++;
    int resul = *r - '0'; r++;
    while (*r >= '0') { resul = resul*10 + *r - '0'; r++; }
    return resul;
  }
  inline void skipNewline() { while (*r != '\n') r++; }
  inline void addPatch(const Pixel &pa) {
    int perfectIndex = pa.r + pa.g*76 + pa.b * 5776;
    if (perfectTable[perfectIndex] == 0 ||
	perfectTable[perfectIndex]->image != numImages) {
      patchNodeVector[newPatchNode].image = numImages;
      patchNodeVector[newPatchNode].next  = perfectTable[perfectIndex];
      perfectTable[perfectIndex]          = patchNodeVector+newPatchNode;
      newPatchNode++;
    }
    perfectTable[perfectIndex]->count++;
  }
  inline void addPatchGeneric(const Pixel &pa);
  void loadImage(char *filename);
  void computeDistances();
  void naiveAPSP();
  void write_result();
};

inline void Contest::addPatchGeneric(const Pixel &pa) {
  const unsigned int cte_hash = 2654435769U;
  int aux   = pa.g ^ (pa.r << 5);
  int index = (aux*cte_hash+pa.b) & hashMask;
  while (hashVector[index].nodes != 0 && hashVector[index].p != pa)
    index = (index + (aux|1)) & hashMask;
  if (hashVector[index].nodes == 0 || hashVector[index].nodes->image != numImages) {
      hashVector[index].p                 = pa;
      patchNodeVector[newPatchNode].image = numImages;
      patchNodeVector[newPatchNode].next  = hashVector[index].nodes;
      hashVector[index].nodes             = patchNodeVector+newPatchNode;
      newPatchNode++;
  }
  hashVector[index].nodes->count++;
}

void Contest::loadImage(char *filename) {
  int    file_descriptor;
  size_t filesize;
  char  *filemapped;
  if ((file_descriptor = open(filename, O_RDONLY)) < 0) {
    fprintf(stderr,"Error opening file \"%s\"\n",filename);
    exit(1);
  }
  struct stat statbuf;
  if (fstat(file_descriptor,&statbuf) < 0) {
    fprintf(stderr,"Error guessing filesize\n");
    exit(1);
  }
  filesize = statbuf.st_size;
  if ((r=filemapped=(char*)mmap(0, filesize,
				PROT_READ, MAP_SHARED,
				file_descriptor, 0))  ==(caddr_t)-1) {
    fprintf(stderr,"Error mmaping\n");
    exit(1);
  }
  //----------------------------------------------------------------------
  skipNewline(); width = readValue();  height = readValue(); readValue();
  if (VERBOSE) fprintf(stderr,"Abriendo fichero %s de %dx%d\n",filename,width,height);
  total_count[numImages] = (width-W+1)*(height-W+1);

  //----------------------------------------------------------------------
  // leer imagen integral: http://en.wikipedia.org/wiki/Summed_area_table
  Pixel *prevRow,*currRow = imagePixels;
  // usamos una matriz de talla height+1 filas, width+1 columnas, la
  // primera fila y la primera columna estan puestas a 0
  for (int j=0; j<=width; ++j) {
    currRow[j].r = currRow[j].g = currRow[j].b = 0;
  }
  for (int i=0; i<height; ++i) {
    prevRow  = currRow;
    currRow += width+1;
    currRow[0].r = currRow[0].g = currRow[0].b = 0;
    //-------------------------------------------------
    int acc_r=0, acc_g=0, acc_b=0;
    for (int j=1; j<=width; ++j) {
      acc_r += readValue()>>bitsq;  currRow[j].r = prevRow[j].r + acc_r;
      acc_g += readValue()>>bitsq;  currRow[j].g = prevRow[j].g + acc_g;
      acc_b += readValue()>>bitsq;  currRow[j].b = prevRow[j].b + acc_b;
    }
  }
  //----------------------------------------------------------------------
  // calcular los patches:
  if (APROXIMO && bitsq == 6 && W == 5) {
    Pixel *row_up   = imagePixels;
    Pixel *row_down = imagePixels+W*(width+1);
    for (int i=0; i<=height-W; ++i) {
      for (int left=0,right=W; left<=width-W; ++left,++right) {
	Pixel pa;
	pa.r = row_down[right].r + row_up[left].r - row_up[right].r - row_down[left].r;
	pa.g = row_down[right].g + row_up[left].g - row_up[right].g - row_down[left].g;
	pa.b = row_down[right].b + row_up[left].b - row_up[right].b - row_down[left].b;
	addPatch(pa);
      }
      row_up   += width+1;
      row_down += width+1;
    }
  } else {
    Pixel *row_up   = imagePixels;
    Pixel *row_down = imagePixels+W*(width+1);
    for (int i=0; i<=height-W; ++i) {
      for (int left=0,right=W; left<=width-W; ++left,++right) {
	Pixel pa;
	pa.r = row_down[right].r + row_up[left].r - row_up[right].r - row_down[left].r;
	pa.g = row_down[right].g + row_up[left].g - row_up[right].g - row_down[left].g;
	pa.b = row_down[right].b + row_up[left].b - row_up[right].b - row_down[left].b;
	addPatchGeneric(pa);
      }
      row_up   += width+1;
      row_down += width+1;
    }
  }
  numImages++;
  munmap(filemapped, filesize); close(file_descriptor);
}

inline int min(int a, int b) {
  return (a<b) ? a : b;
}

void Contest::computeDistances() {
  if (VERBOSE) fprintf(stderr,"computeDistances\n");
  int offset = 0;
  for (int i=0; i<numImages; ++i) { // matriz triangular inferior
    m[i]      = mat+offset;
    m_used[i] = mat_used+offset;
    offset   += i;
  }
  // acumular contadores:
  for (int i=0; i<numImages; ++i)
    for (int j=0; j<i; ++j)
      m[i][j] += total_count[i]+total_count[j];

  if (APROXIMO && bitsq == 6 && W == 5) {
    for (int i=0; i<perfectSize; ++i)
      if (perfectTable[i] && perfectTable[i]->next)
	for (PatchNode *r1=perfectTable[i]; r1!=0; r1=r1->next)
	  for (PatchNode *r2=r1->next; r2!=0; r2=r2->next) {
	    // sabemos que r1->image > r2->image por construccion de la lista
	    m[r1->image][r2->image] -= 2*min(r1->count,r2->count);
	  }
  } else {
    for (int i=0; i<hashSize; ++i)
      if (hashVector[i].nodes && hashVector[i].nodes->next)
	for (PatchNode *r1=hashVector[i].nodes; r1!=0; r1=r1->next)
	  for (PatchNode *r2=r1->next; r2!=0; r2=r2->next) {
	    // sabemos que r1->image > r2->image por construccion de la lista
	    m[r1->image][r2->image] -= 2*min(r1->count,r2->count);
	  }
  }

  // eliminar las que no estan entre las K mejores:
  for (int i=0; i<numImages; ++i) {
    for (int j=0; j<i; ++j) {
      vec_selection[j].ind  = j;
      vec_selection[j].dist = m[i][j];
    }
    for (int j=i+1; j<numImages; ++j) {
      vec_selection[j-1].ind  = j;
      vec_selection[j-1].dist = m[j][i];
    }
    std::nth_element(vec_selection,vec_selection+la_k-1,vec_selection+numImages-1);
    for (int j=0; j<la_k; ++j)
      if (i>vec_selection[j].ind)
  	m_used[i][vec_selection[j].ind] = 1;
      else
  	m_used[vec_selection[j].ind][i] = 1;
  }
}

void Contest::naiveAPSP() {
  // generar listas de aristas y volver a recorrer todo poniendo
  // infinito donde no este marcado:
  int indexWriteEdge = 0;
  for (int i=0; i<numImages; ++i) {
    vertex2edge[i][0] = indexWriteEdge;
    for (int j=0; j<i; ++j)
      if   (m_used[i][j] != 1) { m[i][j] = INFINITE; }
      else                     { edges[indexWriteEdge] = j; ++indexWriteEdge; }
    for (int j=i+1; j<numImages; ++j)
      if (m_used[j][i] == 1)   { edges[indexWriteEdge] = j; ++indexWriteEdge; }
    vertex2edge[i][1] = indexWriteEdge;
  }

  int extras = 0;
  // repetir para cada vertice origen:
  int queueRead, queueWrite;
  for (int u=0; u<numImages; ++u) {
    queueRead  = queueWrite = 0;
    for (int v=0; v<numImages; ++v) // encolar todos los vertices menos orig 
      if (v != u && eme(u,v)<INFINITE) { queue[queueWrite] = v; stamp[v] = true; ++queueWrite; }
    for (;queueRead < queueWrite; ++queueRead) { //mientras no se agote la cola:
      int v  = queue[queueRead], uv = eme(u,v);
      for (int edgeIndex = vertex2edge[v][0]; edgeIndex < vertex2edge[v][1]; ++edgeIndex) {
	int w = edges[edgeIndex];
	if (w <= u) { vertex2edge[v][0]=edgeIndex+1; }
	else {
	  int uvw = uv + eme(v,w);
	  int &uw = m[w][u];
	  if (uvw < uw) {
	    uw = uvw;
	    if (!stamp[w]) { 
	      queue[queueWrite] = w; stamp[w] = true; ++queueWrite;
	    }
	  }
	}
      }
      stamp[v] = false;
    }
  }
  fprintf(stderr,"extras = %d\n",extras);
}

void Contest::write_result() {
  //----------------------------------------------------------------
  // IMPRIMIR RESULTADO:
  for (int i=0; i<numImages; ++i) {
    for (int j=0; j<numImages; ++j) {
      int value = 0;
      if (i>j) value = m[i][j];
      if (i<j) value = m[j][i];
      if (value == INFINITE) printf("inf");
      else                   printf("%d",value);
      if (j<numImages-1) putchar(' ');
    }
    putchar('\n');
  }
}

//----------------------------------------------------------------------
Contest contest; // global
int main(int argc,char *argv[],char *env[]) {
  if (argc<5) { fprintf(stderr,"concurso file k w q\n"); exit(1); }
  contest.la_k = atoi(argv[2]);
  contest.W    = atoi(argv[3]);
  int q        = atoi(argv[4]);
  contest.bitsq= 0; while (q > 1) { q = q >> 1; contest.bitsq++; }
  FILE *fe=fopen(argv[1],"rt");
  if (fe==NULL) {
    fprintf(stderr,"no se puede abrir fichero %s\n",argv[1]);
    exit(1);
  }
  char filename[1024];
  while(fscanf(fe,"%s",filename) == 1)
    contest.loadImage(filename);

  contest.computeDistances();
  contest.naiveAPSP();
  contest.write_result();

  return 0;
}
